#!/usr/bin/perl

# index2html.pl
#    Programmed by Pasi Ojala <p.ojala@iki.fi>
#    Enhanced by Marko Mäkelä <marko.makela@iki.fi>
# Last revision: June 1, 2005

# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.

# This script reads in file descriptions from a 00INDEX file, adds any new
# files and removes any removed files from the 00INDEX file.  Finally it
# generates an index.html file with a directory listing, including metadata.
#
# Standard output is used for messages that let you know what is happening;
# they can be safely ignored.  These messages will not be printed, unless the
# -v (verbose) option is being used.
#
# The output from standard error should be read by the user.
#
# Format of the 00INDEX file:
#
#--8<--example--8<--example--8<--example--8<--example--8<--example--8<--
# ** A global directory description is marked with two asterisks
# ** at beginning of the file
# directory1/
#   Description of directory1.
#   This can consist of several lines.
# directory2/
#   Description of directory2
# .../
# last_directory/
#   The descriptions are not obligatory.
#
# file1
#   Description of file1
#   may consist of several lines
# file2-part1of2.sfx
# file2-part2of2.sfx
#   A description may apply to several files.
#--8<--example--8<--example--8<--example--8<--example--8<--example--8<--
#
# The rendered HTML version most probably have implicit line breaks in
# different positions, depending on the view area width.  Sometimes you
# may want to force a line break in index.html.  To do this, add a space
# (or any whitespace character) to the line that should end in a line break.
# In the following example, the trailing spaces have been converted to
# underscores ("_") to emphasize them.	You must use spaces in the
# actual 00INDEX file; the underscores won't work!
#
# Alternatively, you may leave blank lines in the descriptions.
#
# Note that no indentations with whitespaces will be produced in the
# HTML output.  You must do indentations with e.g. "-" or "*",
# as in the following example.
#
#--8<--example--8<--example--8<--example--8<--example--8<--example--8<--
# ** Welcome to the Commodore 8-bit file area
# ** at ftp.funet.fi:/pub/cbm!
# **
# ** This site is located in Finland, Europe.
# ** Other major Commodore sites include:_
# ** - ftp.giga.or.at:/pub/c64 (Austria, Europe)_
# ** - ccnga.uwaterloo.ca (Canada, North America)
#--8<--example--8<--example--8<--example--8<--example--8<--example--8<--
#
# The forced line breaks work with all descriptions (global directory
# description, directory descriptions and file descriptions).
#
# This script creates links for URLs that match the regular expression
# $URLexpression after the HTML conversion.
#
# If there is a $index file but no 00INDEX file in a
# directory, the directory will be skipped.

$recursive = '';
$verbose = '';
$overwrite = '';
%fileindex = ();
%sizeindex = ();
%dateindex = ();
$comments = '';
$index = "index.html";

#URL expression match pattern
$URLexpression = '&lt;URL:(.+?)&gt;';

while($#ARGV >= 0)
{
    $opt = shift(@ARGV);
    if($opt eq '-R')
    {
	$recursive = 'T';
    }
    elsif($opt eq '-v')
    {
	$verbose ='T';
    }
    elsif($opt eq '-w')
    {
	$overwrite = 'T';
    }
    else
    {
	print STDERR "Usage: $0 [-R] [-v] [-w]\n";
	exit(10);
    }
}

# try to find out where we are.
$curdir = `pwd`;
chop($curdir) if($curdir =~ /[\n\r]$/);
$ftpdir = $curdir . '/';

if($ftpdir =~ /pub\/(.*)$/)
{
    $ftpdir = "/pub/$1";
}
else
{
    print STDERR "Warning: the pathnames will not be converted!\n";
}

if (($ftpdir =~ tr|/|/|) > 3)
{
    die unless opendir (DIR, "$curdir/..");
    my @dirs = sort (grep
		     { !/^(\..*|incoming)$/ &&
			   (! -l "$curdir/../$_") &&
			   (-d "$curdir/../$_") }
		     readdir (DIR));
    closedir (DIR);
    my $diridx;
    $_ = substr ($curdir, 1 + rindex ($curdir, '/'));
    for ($diridx = 0; $diridx <= $#dirs; $diridx++)
    {
	last if $dirs[$diridx] eq $_;
    }
    die "directory $curdir/../$_ not found\n" if ($diridx > $#dirs);

    &processdir($ftpdir, $curdir,
		$diridx > 0 ? $dirs[$diridx - 1] : undef,
		$diridx < $#dirs ? $dirs[$diridx + 1] : undef);
}
else
{
    &processdir($ftpdir, $curdir, undef, undef);
}
exit(0);

# replace a file if it has changed
sub replace
{
    my ($new, $old) = @_;
    $currentdir = `pwd`;
    if (system ("cmp", "-s", $old, $new) >> 8)
    {
	die "could not create $currentdir/$old from $new: $!\n" if
	    system ("mv", $new, $old) >> 8;
	chmod 0664, $old;
    }
    else
    {
	unlink $new;
    }
}

sub dirworthy($){
    my $string = shift;
    if ( ($string =~ m/d\d\d\.gz$/) || ($string =~ m/d\d\d$/) ) {
        return 1;
    }
    if ( ($string =~ m/D\d\d\.gz$/) || ($string =~ m/D\d\d$/) ) {
        return 1;
    }
    if ( ($string =~ m/t\d\d\.gz$/) || ($string =~ m/t\d\d$/) ) {
        return 1;
    }
    if ( ($string =~ m/T64.gz$/) || ($string =~ m/T64$/) ) {
        return 1;
    }
    if ( ($string =~ m/d\d\d\.zip$/) || ($string =~ m/D\d\d.zip$/) ) {
        return 1;
    }
}

sub dirlinkapp($) {
    my $string = shift;
    if ( dirworthy( $string ) ) {
        return '&nbsp;&nbsp;&nbsp;&nbsp;<a href="/cgi-bin/d64mod.cgi?' 
        . &htmlname($cwd)
        . $string . '"><font color=black>(<i>dir</i>)</font></a>';
    }
    else {
        return '';
    }
}

sub processdir
{
    local(%dirindex) = ();
    %fileindex = ();
    %sizeindex = ();
    %dateindex = ();
    $comments = '';
    local($cwd, $pwd, $prev, $next) = @_;

    chdir $pwd || die "cannot chdir to $pwd\n";

    if(open(INDEX, '00INDEX'))
    {
	$name = '';
	$info = '';
	while(<INDEX>)
	{
	    chop while(/[\n\r]$/);
	    if(/^\*\*.*$/)
	    {
		$comments .= $_ . "\n";
	    }
	    else
	    {
		if(/^(\S.*)\s*$/)
		{
		    if($name ne '')
		    {
			&addentry($name, $info);
		    }
		    $name = $1;
		    $info = "\n";
		}
		elsif(/^\s+(\S.*)$/)
		{
		    if($info eq "\n")
		    {
			$info = '  ' . $1 . "\n";
		    }
		    else
		    {
			$info .= '  ' . $1 . "\n";
		    }
		}
		elsif ($info !~ /\n\n$/)
		{
		    $info .= "\n";
		}
	    }
	}
	if($name ne '')
	{
	    &addentry($name, $info);
	}
	close(INDEX);
    }
    elsif (-r $index)
    {
	if ($verbose)
	{
	    print STDOUT "skipping directory $cwd: no 00INDEX file\n";
        }

	return;
    }

    if(opendir(DIR, "."))
    {
	@allfiles = grep(!/^\./, readdir(DIR));
	closedir(DIR);
    }
    for(@allfiles)
    {
	my($dev,$ino,$mode)=stat($_);

	if ((-l $_) || ($mode & 4)) # only process world-readable files
	{
	    if(-d $_)
	    {
		if(!$dirindex{"$_/"})
		{
		    print STDERR "Directory $cwd$_/ missing from 00INDEX!\n";
		    $dirindex{"$_/"} = "";
		}
		if ((~$mode) & 0555)
		{
		    print STDERR "Warning: some r or x bits of $cwd$_/ not set!\n";
		}
	    }
	    elsif (!($mode & 4))
	    {
		print STDERR "File $cwd$_ is a symlink pointing to nowhere!\n";
	    }
	    else
	    {
		if (!$fileindex{$_} && ($_ ne '00INDEX') &&
		    ($_ ne $index) && !(/\~$/))
		{
		    print STDERR "File $cwd$_ missing from 00INDEX!\n";
		    $fileindex{$_} = "";
		    $sizeindex{$_} = $size;
		    $dateindex{$_} = $mtime;
		}
	    }
	}
    }

    if($overwrite || ! (-f '00INDEX'))
    {
	if(-f '00INDEX')
	{
	    rename('00INDEX', '00INDEX~');
	    chmod 0660, '00INDEX~';
	}

	if(open(INDEX, '>00INDEX'))
	{
	    if ($comments)
	    {
		print INDEX $comments, "\n";
	    }

	    @keys = sort { "\L$a" cmp "\L$b" } (keys(%dirindex));

# remove newlines from the last directory description
	    if ($#keys >= 0)
	    {
		$dirindex{@keys[$#keys]} =~ s/\n+$/\n/g;
	    }

	    for(@keys)
	    {
		$tmp = $dirindex{$_};
		if($tmp eq "\n")
		{
		    print INDEX "$_\n";
		}
		else
		{
		    print INDEX "$_\n$tmp";
		}
	    }

	    if ($#keys >= 0)
	    {
		print INDEX "\n";
	    }

	    for(sort { "\L$a" cmp "\L$b" } (keys(%fileindex)))
	    {
		if($fileindex{$_} eq "\n")
		{
		    print INDEX "$_\n";
		}
		else
		{
		    print INDEX "$_\n$fileindex{$_}";
		}
	    }
	    close(INDEX);
	    chmod 0664, '00INDEX';
	}
	else
	{
	    print STDERR "Could not recreate $pwd/00INDEX!\n";
	    rename('00INDEX~', '00INDEX');
	    chmod 0664, '00INDEX';
	}
    }

    open (HTML, ">/tmp/$index$$") || warn "Could not create /tmp/$index$$!\n";

    print HTML '<?xml version="1.0"?><!-- -*- xml -*- -->
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
 "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html
 xmlns="http://www.w3.org/1999/xhtml"
 lang="en" xml:lang="en" dir="ltr">

<head>
<title>', &htmlname($cwd), '</title>
<link rel="stylesheet" href="', ('../' x (($cwd =~ tr|/|/|) - 3)),
'.index.css" />
<link rel="icon" href="', ('../' x (($cwd =~ tr|/|/|) - 3)),
'.icon.png" />
<link rel="SHORTCUT ICON" href="', ('../' x (($cwd =~ tr|/|/|) - 3)),
'.icon.ico" />
';
    if (($cwd =~ tr|/|/|) > 3)
    {
	print HTML "<link rel=\"top\" href=\"../index.html\" />\n";
    }
#   else
#   {
#	print HTML "<link rel=\"top\" href=\"../\" />\n";
#   }
    print HTML "<link rel=\"prev\" href=\"../",
	&urlname($prev), "/index.html\" />\n" if defined ($prev);
    print HTML "<link rel=\"next\" href=\"../",
	&urlname($next), "/index.html\" />\n" if defined ($next);
    print HTML
'<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
</head><body>
<img src="/zimmersfiles2.png" ALIGN=LEFT ALT="Zimmers.NET Files (formerly ftp.funet.fi)" HEIGHT=74 WIDTH=130 style="position:absolute;left:0px">
<h1 style="position: relative">', &dirlinks($cwd, $index), '</h1>
<BR CLEAR=LEFT>
<div class="top">
';

#Replace the <, > and & signs with their HTML equivalents.
    $tmp = &htmlname($comments);
#Generate a <br /> tag for a newline only for plain "**" lines
#or if the previous line ends in one or more whitespaces.
    $tmp =~ s|\n\*\*\n|<br />\n|g;
    $tmp =~ s|\s+\n\*\*\s*|<br />\n|g;
#Strip off the '**'s and white space at line beginnigns.
    $tmp =~ s/^\*\*\s*//g;
    $tmp =~ s/\n\*\*\s*/\n/g;
#Strip off trailing newlines.
    $tmp =~ s/\n+$//;

#Generate links for URLs
    $tmp =~ s/$URLexpression/<a class="link" href="\1">\1<\/a>/go;

    if($tmp ne '')
    {
	print HTML "<p>$tmp</p>\n";
    }

    @keys = sort{ "\L$a" cmp "\L$b" } (keys(%dirindex));

    print HTML "</div>\n";
    print HTML '<h2><a class="anchor" name="dirs">Subdirectories</a></h2>
<div class="dirs">
<dl>
' if $#keys >= 0;

    for(@keys)
    {
	$tmp = $_;
	chop $tmp;

	if ($tmp ne "incoming" && -l $tmp)
	{
	    $tmp2 = &resolvelink($tmp,$pwd);
	    print HTML "<dt><a href\=\"",
		       &urlname($tmp2) . "/" .
				&indexname($index, $tmp2),
		       "\">", &htmlname($tmp), "\@</a></dt>\n";
	}
	else
	{
	    print HTML "<dt><a href\=\"",
		       &urlname($tmp . "/" .
				&indexname($index, $tmp)),
		       "\">", &htmlname($tmp), "/</a></dt>\n";
	}

#Replace the <, > and & signs with their HTML equivalents in the description.
	$tmp = &htmlname($dirindex{$_});
#Strip off any leading whitespaces and trailing newlines in the comments.
#Generate a <br /> tag for a newline only if
#the previous line ends in one or more whitespaces.

	$tmp =~ s/\n+$//;
	$tmp =~ s/^\s*//;
	$tmp =~ s|\s+\n\s*|<br />\n|g;
	$tmp =~ s/\n\s*/\n/g;

#Generate links for URLs
	$tmp =~ s/$URLexpression/<a class="link" href="\1">\1<\/a>/go;

	if($tmp)
	{
	    print HTML "<dd>$tmp</dd>\n";
	}
    }

    print HTML "</dl>\n</div>\n" if $#keys >= 0;

    @keys = sort { "\L$a" cmp "\L$b" } (keys(%fileindex));
    print HTML '<h2><a class="anchor" name="files">Files</a></h2>
<div class="files">
<dl>
' if $#keys >= 0;

    for(@keys)
    {
	$tmp = $_;

	if (-l $tmp)
	{
	    $tmp2 = &resolvelink($tmp,$pwd);
	    print HTML "<dt><a href\=\"", &urlname($tmp2), "\">",
		       &htmlname($tmp), "\@</a>", &dirlinkapp($tmp),
               "\n", &filedate($tmp),
		       "\n<span class=\"length\">",
		       $sizeindex{$tmp}, "</span></dt>\n";
	}
	else
	{
	    print HTML "<dt><a href=\"", &urlname($tmp), "\">",
		       &htmlname($tmp), "</a>", &dirlinkapp($tmp),
               "\n", &filedate($tmp),
		       "\n<span class=\"length\">",
		       $sizeindex{$tmp}, "</span></dt>\n";
	}

#Replace the <, > and & signs with their HTML equivalents in the description.
	$tmp = &htmlname($fileindex{$_});
#Strip off any leading whitespaces and trailing newlines in the comments.
#Generate a <br /> tag for a newline only if
#the previous line ends in one or more whitespaces.
	$tmp =~ s|\n\n\n*|<br /><br />|g;
	$tmp =~ s/^\s*//;
	$tmp =~ s/\n+$//;
	$tmp =~ s|\s+\n\s*|<br />\n|g;
	$tmp =~ s|\n\s*|<br />\n|g;

#Generate links for URLs
	$tmp =~ s/$URLexpression/<a class="link" href="\1">\1<\/a>/go;

	if($tmp)
	{
	    print HTML "<dd>$tmp</dd>\n";
	}
	else
	{
	    #print HTML "<dd>&nbsp;</dd>\n";
	}
    }

    print HTML "</dl>\n</div>\n" if $#keys >= 0;
    my($uri) = '../' x (($cwd =~ tr|/|/|) - 3);
    print HTML
"<p class=\"bot\"><a href=\"${uri}MIRRORS.html\">Mirror sites</a>\n&ndash;
<a href=\"${uri}README.html\">General information</a>\n&ndash;
<a href=\"${uri}FILETYPES.html\">File types</a>";
    print HTML "\n&ndash;
<a href=\"${uri}crossplatform/transfer/transfer.html\">Data transfer</a>"
	if $cwd =~ m|/cbm/|;
    print HTML "</p>\n";
    print HTML "</body></html>\n";

    close (HTML);
    &replace ("/tmp/$index$$", $index);

    if($recursive)
    {
	my @dirs = sort (grep { $_ ne "incoming/" && (! -l $_) }
			 map { s/\/+$//so; $_ } keys(%dirindex));
	my $diridx;
	for ($diridx = 0; $diridx <= $#dirs; $diridx++)
	{
	    $_ = $dirs[$diridx];
	    if (defined ($_))
	    {
		print STDOUT "Processing $cwd$_\n" if $verbose;
		&processdir("$cwd$_/", "$pwd/$_",
			    $diridx > 0
			    ? $dirs[$diridx - 1] : undef,
			    $diridx < $#dirs
			    ? $dirs[$diridx + 1] : undef);
		chdir $pwd || die "Can't cd back to $pwd!\n";
	    }
	}
    }
}

sub addentry
{
    local($name, $info) = @_;

#remove the nonsense descriptions "[no description]" and "--"
#   $info =~ s/^\s*(\[no description\]|--)\s*\n+$/\n\n/;

#remove initial newlines
    $info =~ s/^\n+(.)/\1/g;

#if there are more than 2 successive newlines, truncate them to 2
    $info =~ s/\n\n\n+/\n\n/g;

#remove comments consisting of only newlines
    $info =~ s/^\n\n+$/\n\n/gso;

    ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime)=stat($name);

    if (!($mode & 4))
    {
	if(-d $name)
	{
	    print STDERR "Dir $cwd$name can't be read by everyone\n";
	}
	else
	{
	    print STDERR "File $cwd$name can't be read by everyone\n";
	}
    }
    elsif ($name =~ /\/$/)
    {
	# directory
	if(-d $name)
	{
	    $dirindex{$name} = $info;
	}
	else
	{
	    print STDERR "Dir $cwd$name does not exist\n";
	}
    }
    else
    {
	# file
	if(-f $name)
	{
	    $fileindex{$name} = $info;
	    $sizeindex{$name} = $size;
	    $dateindex{$name} = $mtime;
	}
	else
	{
	    print STDERR "File $cwd$name does not exist\n";
	}
    }
}

sub urlname
{
    local($name) = @_;
    $name =~ s/([\000-\040\"\#%\?:+&<>\176-\377])/"%".unpack("H2",$1)/ges;
    return $name;
}

sub htmlname
{
    local($name) = @_;
#Replace the <, > and & signs with their HTML equivalents.
    $name =~ s/&/&amp\;/g;
    $name =~ s/</&lt\;/g;
    $name =~ s/>/&gt\;/g;
    return $name;
}

sub indexname
{
    local($name, $dir) = @_;
    local($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime) =
	stat ($dir.'/'.$name);
    return $name if ($mode & 4);
    for ("index.html","index.htm")
    {
	($dev,$ino,$mode) = stat ($dir.'/'.$_);
	return $_ if ($mode & 4);
    }
    return '';
}

sub dirlinks
{
    my($path, $indexfile) = @_;
    my(@dirs) = split("/", $path);
    my($retval) = $path = '/';
    my($dir) = '';
    shift @dirs;
    $retval .= &htmlname (shift @dirs) . "/";
    while(defined($dir = shift @dirs))
    {
	$path .= $dir . "/";
	if ($#dirs >= 0)
	{
	    $retval .= "<a\nhref=\"" . &urlname("../" x ($#dirs + 1) .
		    &indexname($indexfile, $cwd . ("../" x ($#dirs + 1)))) .
		    "\">" . &htmlname($dir) . "</a>/";
	}
	else
	{
	    $retval .= &htmlname($dir) . "/";
	}
    }

    $retval;
}

sub filedate
{
    my($name) = @_;
    my($timestamp) = $dateindex{$name};
    my($sec,$min,$hour,$mday,$mon,$year) = gmtime($timestamp);
    my($class) = time - $timestamp < 2592000 # older than one month
	? "recentdate"
	: "date";
    return "<span class=\"$class\">" .
	sprintf("%04u-%02u-%02u", 1900 + $year, $mon + 1, $mday) .
	"</span>";
}

sub resolvelink
{
    my($name,$pwd,$subdir) = @_;
    $workdir = $pwd;
    $name = res_link($name,$pwd);
    if ($name !~ /\// && (-f $name))
    {
	my($dir) = `pwd`; chop $dir;
	$name = &truncpath($dir.'/'.$name,$workdir);
    }
    chdir $pwd || die "could not chdir back to `$pwd'\n";
    return $name;
}

sub res_link
{
    my($name,$pwd) = @_;
    if (!(-l $name))
    {
	(-f $name) || warn "res_link called for `$name' in `$pwd'\n";
	return $name;
    }
    $name = readlink($name);
    if (-f $name)
    {
	my($dir) = $name;
	$dir =~ s|/([^/]*)$|| || return &res_link($name,$pwd);
	($dir ne "") || return &res_link($1,$pwd);
	my($file) = $1;
	chdir $dir || die "could not chdir to `$dir'\n";
	$dir=`pwd`;chop $dir;
	$name = &truncpath($dir,$workdir).'/'.&res_link($file,$dir);
	return $name;
    }
    elsif (-d $name)
    {
	chdir $name || die "could not chdir to `$name'\n";
	$name=`pwd`;chop $name;
	return &truncpath($name,$workdir);
    }
    else
    {
	warn "`$name' neither file or directory\n";
	return $name;
    }
}

sub truncpath ()
{
    my($path,$prefix) = @_;
    $path =~ s|^\Q$prefix\E/|| && return $path;

    my($dir) = $path;
    while ($prefix !~ s|^\Q$dir/\E||)
    {
	$dir =~ s|/[^/]*$|| ||
	    die "truncpath (`$path', `$prefix') failed\n";
    }
    $path =~ s|^\Q$dir/\E|| || $path =~ s|^\Q$dir\E$|| ||
	die "could not truncate `$path' with `$dir'\n";
    for (split ('/', $prefix))
    {
	$path = '../' . $path;
    }
    $path =~ s|/$||;

    return $path;
}
